<?php

// $psp_id sn_dns.php date 20150911

define("RR_A",     1); // host address
define("RR_NS",    2); // authoritative name server
define("RR_CNAME", 5); // canonical name for alias
define("RR_MX",   15); // mail exchange
define("RR_AAAA", 28); // IPv6 address

$sn_dns_pid = 0;
$sn_dns_tid = 0;
$sn_dns_query_type = 0;
$sn_dns_next_tick = 0;

function sn_dns_get_tick()
{
	while(($pid = pid_open("/mmap/st9", O_NODIE)) == -EBUSY)
		usleep(500);

	if(!pid_ioctl($pid, "get state"))
		pid_ioctl($pid, "start");

	$tick = pid_ioctl($pid, "get count");
	pid_close($pid);

	return $tick;
}

function sn_dns_encode_name($name)
{
	$enc = "";

	while($name)
	{
		$len = strpos($name, ".");

		if($len === false)
		{
			$enc .= int2bin(strlen($name), 1);
			$enc .= $name;
			break;
		}
		else
		if($len > 0)
		{
			$enc .= int2bin($len, 1);
			$enc .= substr($name, 0, $len);
			$name = substr($name, $len + 1);
		}
		else
		{
			echo "sn_dns: invalid RR name $name\r\n";
			return "";
		}
	}

	$enc .= "\x00";

	return $enc;
}

function sn_dns_decode_name($name, &$rbuf)
{
	$dec = "";

	for(;;)
	{
		$len = bin2int($name, 0, 1);

		if($len & 0xc0)
		{
			$len = bin2int($name, 0, 2, true) & 0x3fff;
			$name = substr($rbuf, $len);
			continue;
		}

		$dec .= substr($name, 1, $len);
		$name = substr($name, 1 + $len);

		if(bin2int($name, 0, 1))
			$dec .= ".";
		else
			return $dec;
	}
}

function sn_dns_skip_name(&$name)
{
	do
	{
		$len = bin2int($name, 0, 1);

		if($len & 0xc0)
		{
			$name = substr($name, 2);
			return;
		}

		if((1 + $len) > strlen($name))
			$len = strlen($name) - 1;

		$name = substr($name, $len + 1);
	}while($len && $name);
}

function dns_setup($udp_id, $server_addr = "")
{
	global $sn_dns_pid, $sn_dns_tid;

	if(!$server_addr)
	{
		if((int)ini_get("init_net0"))
			$pid_net = pid_open("/mmap/net0");
		else
			$pid_net = pid_open("/mmap/net1");
		$server_addr = pid_ioctl($pid_net, "get nsaddr");
		pid_close($pid_net);
	}

	if(!$sn_dns_pid)
	{
		$sn_dns_pid = pid_open("/mmap/udp$udp_id");
		$sn_dns_tid = rand(0, 65536);
	}

	pid_ioctl($sn_dns_pid, "set dstaddr $server_addr");
	pid_ioctl($sn_dns_pid, "set dstport 53");
}

function dns_send_query($name, $type)
{
	global $sn_dns_query_type, $sn_dns_next_tick;
	global $sn_dns_pid, $sn_dns_tid;

	if(!$sn_dns_pid)
		dns_setup(0, "");

	$sn_dns_query_type = $type;
	$sn_dns_next_tick = sn_dns_get_tick() + 2000; // 2 seconds

	$sn_dns_tid++;
	$query  = int2bin($sn_dns_tid, 2, true);
	$query .= hex2bin("0100"); // QR(0), OPCODE(0), AA(0), TC(0), RD(1), RA(0), Z(0), RCODE(0)
	$query .= hex2bin("0001"); // question count 1
	$query .= hex2bin("0000"); // answer count 0
	$query .= hex2bin("0000"); // name server count 0
	$query .= hex2bin("0000"); // additional count 0
	$query .= sn_dns_encode_name($name);
	$query .= int2bin($type, 2, true); // qtype
	$query .= hex2bin("0001"); // class : IN

	pid_sendto($sn_dns_pid, $query);
}

function dns_loop()
{
	global $sn_dns_query_type, $sn_dns_next_tick;
	global $sn_dns_pid, $sn_dns_tid;

	if(sn_dns_get_tick() >= $sn_dns_next_tick)
	{
		echo "sn_dns: lookup timeout\r\n";
		$sn_dns_next_tick = 0;
	}

	if(!$sn_dns_next_tick)
		return "";

	if(!pid_ioctl($sn_dns_pid, "get rxlen"))
		return false;

	$sn_dns_next_tick = 0;

	$rbuf = "";

	pid_recvfrom($sn_dns_pid, $rbuf);

	$questions = bin2int($rbuf, 4, 2, true);
	$answers = bin2int($rbuf, 6, 2, true);

	if(!$answers)
		return "";

	$pbuf = substr($rbuf, 12); // skip dns header

	for($i = 0; $i < $questions; $i++)
	{
		sn_dns_skip_name($pbuf);  // skip QNAME
		$pbuf = substr($pbuf, 4); // skip QTYPE/QCLASS
	}

	for($i = 0; $i <= $answers; $i++)
	{
		sn_dns_skip_name($pbuf);  // skip NAME

		$rr_type = bin2int($pbuf, 0, 2, true); 
		$pbuf = substr($pbuf, 4); // skip TYPE/CLASS
		$pbuf = substr($pbuf, 4); // skip TTL
		$rd_len = bin2int($pbuf, 0, 2, true);
		$pbuf = substr($pbuf, 2); // skip RDLENGTH

		if($rr_type == $sn_dns_query_type)
		{
			switch($sn_dns_query_type)
			{
				case RR_A:
					return inet_ntop(substr($pbuf, 0, 4));

				case RR_NS:
					return sn_dns_decode_name($pbuf, $rbuf);

				case RR_MX:
					$pbuf = substr($pbuf, 2); // skip PREFERENCE
					return sn_dns_decode_name($pbuf, $rbuf);
			}
		}

		$pbuf = substr($pbuf, $rd_len); // skip RDATA
	}

	return "";
}

function dns_lookup($name, $type)
{
	dns_send_query($name, $type);

	while(1)
	{
		$rr = dns_loop();

		if($rr === false)
			usleep(1000);
		else
		{
			if($rr == "")
				return $name;
			else
				return $rr;
		}
	}
}

?>
