<?php

// $psp_id sn_esmtp.php date 20150917

///////////////////////////////////////////////////////////////////////////
//
// sn_mysql.php
//
// Description:
//  MySQL is the world's second most widely used open-source relational
//  database management system. The MySQL development project has made its
//  source code available under the terms of the GNU Green Public License,
//  as well as under a variety of proprietary agreements.
//  MySQL is a popular choice of database for use in web applications,
//  and is a central component of the widely used LAMP(Linux, Apache, MySQL, Perl/PHP/Python)
//  open source web application software stack (and other 'APM' stacks).
//
// Revision: 1.4 (2015.08.24)
//
// Copyright: Sollae Systems Co., Ltd. 2014
//
// License:
//  This PHPoC code is open source that is licensed so that others
//  are free to use, change, and distribute this codes.
//
// Exemption from Liability:
//  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
//  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
//  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
//  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
//  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
//  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
//  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
//  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
//  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
//  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//  
//////////////////////////////////////////////////////////////////////////

//global variables
$mysql_pid = 0;
$mysql_state = 0;
$mysql_next_tick = 0;
$mysql_server_addr = "";
$mysql_user_name = "";
$mysql_raw_pwd = "";
$mysql_query_str = "";
$mysql_db_name = "";
$mysql_fetch_count = 0;
$mysql_text_pointer = 0;

//constants
define("MYSQL_STATE_IDLE",        0);
define("MYSQL_STATE_READY",       1);
define("MYSQL_STATE_CON",         2);
define("MYSQL_STATE_CON_FIN",     3);
define("MYSQL_STATE_AUTH",        4);
define("MYSQL_STATE_QUERY_READY", 5);
define("MYSQL_STATE_QUERY",       6);

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_get_tick 
// Description   : get current system tick
// Parameters    : none
// Return Values : $tick - current system tick
//////////////////////////////////////////////////////////////////////////
function mysql_get_tick()
{
	while(($pid = pid_open("/mmap/st9", O_NODIE)) == -EBUSY)
		usleep(500);
	if(!pid_ioctl($pid, "get state"))
		pid_ioctl($pid, "start");
	$tick = pid_ioctl($pid, "get count");
	pid_close($pid);
	return $tick;
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_hash
// Description   : hash messages by SHA1
// Parameters    : $msg - a string to be hashed
// Return Values : hashed data
//////////////////////////////////////////////////////////////////////////
function mysql_hash($msg)
{
	if(PHP_VERSION_ID > 10000)
		return hash("sha1", $msg, true);
	else
	{
		$sha1 = system("sha1 init");
		system("sha1 update %1 %2", $sha1, $msg);
		return system("sha1 final %1", $sha1);
	}
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_make_pwd 
// Description   : make MySQL hashed password with raw password
// Parameters    : $rdata - received random data from DB server
//               : $raw_pwd - raw password
// Return Values : $data - hashed password
//////////////////////////////////////////////////////////////////////////
function mysql_make_pwd($rdata, $raw_pwd)
{
	$data = "";
	$data1 = mysql_hash($raw_pwd);
	$data2 = mysql_hash($data1);
	$sdata = $rdata . $data2;
	$data3 = mysql_hash($sdata);
	for($i = 0; $i < 5; $i++)
	{
		$data_int1 = bin2int($data1, $i*4, 4);
		$data_int2 = bin2int($data3, $i*4, 4);
		$data_int = $data_int1 ^ $data_int2;
		$data .= int2bin($data_int, 4);
	}
	return $data;
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_close
// Description   : close MySQL connection
// Parameters    : $tcp_id - tcp id of MySQL connection
// Return Values : false - on failure
//               : true - on success
//////////////////////////////////////////////////////////////////////////
function mysql_close($tcp_id)
{
	global $mysql_state;

	if(pid_ioctl($tcp_id, "get state") != TCP_CONNECTED)
		return false;
	else
	{
		pid_send($tcp_id, hex2bin("0100000001"));
		$mysql_state = MYSQL_STATE_IDLE;
		pid_close($tcp_id);
		return true;
	}
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_init
// Description   : initialize MySQL parameters
// Parameters    : $server - IP address of DB Server
//               : $user_name - user account ID
//               : $raw_pwd - user account password
// Return Values : none
//////////////////////////////////////////////////////////////////////////
function mysql_init($server, $user_name, $raw_pwd)
{
	global $mysql_state;
	global $mysql_next_tick;
	global $mysql_pid;
	global $mysql_server_addr, $mysql_user_name, $mysql_raw_pwd;

	$mysql_server_addr = $server;
	$mysql_user_name = $user_name;
	$mysql_raw_pwd = $raw_pwd;

	$mysql_state = MYSQL_STATE_READY;
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_con_loop()
// Description   : make progress MySQL connection phase in a loop
// Parameters    : $tcp_id - tcp id for $mysql_pid
// Return Values : $mysql_pid - pid of MySQL connection
//               : false - for next call
//               : -1 - on failure
//////////////////////////////////////////////////////////////////////////
function mysql_con_loop($tcp_id)
{
	global $mysql_pid;
	global $mysql_state;
	global $mysql_next_tick;
	global $mysql_server_addr, $mysql_user_name, $mysql_raw_pwd;

	if($mysql_state == MYSQL_STATE_IDLE)
		return -1;

	$rbuf = "";

	if($mysql_state > MYSQL_STATE_CON)
	{
		$len = pid_ioctl($mysql_pid, "get rxlen");
		if(!$len)
		{
			$state = pid_ioctl($mysql_pid, "get state");
			if($state != TCP_CONNECTED)
			{
				echo "mysql: connection closed!\r\n";
				return -1;
			}
			if(mysql_get_tick() >= $mysql_next_tick)
			{
				echo "mysql: receve timeout!\r\n";
				mysql_close($mysql_pid);
			}
			return false;
		}
		pid_recv($mysql_pid, $rbuf, $len);
	}

	switch($mysql_state)
	{
	case MYSQL_STATE_READY:
		$pos = strpos($mysql_server_addr, ':');
		if($pos !== false)
		{
			$port = (int)(substr($mysql_server_addr, ($pos + 1)));
			$mysql_server_addr = substr($mysql_server_addr, 0, ($pos));
		}
		echo "mysql: connect to $mysql_server_addr : 3306...\r\n";
		$mysql_pid = pid_open("/mmap/tcp$tcp_id");
		pid_connect($mysql_pid, $mysql_server_addr, 3306);
		$mysql_next_tick = mysql_get_tick() + 5000;
		$mysql_state = MYSQL_STATE_CON;
		break;

	case MYSQL_STATE_CON:
		$state = pid_ioctl($mysql_pid, "get state");
		if($state == TCP_CONNECTED)
		{
			echo "mysql: connection completed!\r\n";
			$mysql_state = MYSQL_STATE_CON_FIN;
			return false;
		}
		if(mysql_get_tick() >= $mysql_next_tick)
		{
			echo "mysql: connection timeout!\r\n";
			return -1;
		}
		break;

	case MYSQL_STATE_CON_FIN:
		$rlen = bin2int($rbuf, 0, 3);
		$rbuf = substr($rbuf, 4, $rlen);
		$version_ref = strpos($rbuf, "\x00");

		//$version = substr($rbuf, 0, $version_ref);

		$salt_ref_1 = $version_ref + 5;
		$salt_ref_2 = $salt_ref_1 + 27;
		$salt = substr($rbuf, $salt_ref_1, 8) . substr($rbuf, $salt_ref_2, 12);
		
		$pwd = mysql_make_pwd($salt, $mysql_raw_pwd);

		$sdata_body = hex2bin("85a2");          //[ 2]Client Capabilities
		$sdata_body .= hex2bin("0e00");         //[ 2]Extended Client Capabilities
		$sdata_body .= hex2bin("00000040");     //[ 4]Max Packet
		$sdata_body .= hex2bin("08");           //[ 1]Charset
		for($i = 0; $i < 23; $i++)              //[23]Reserved 23 Bytes
			$sdata_body .= "\x00";
		$sdata_body .= $mysql_user_name;        //[ ?]Username
		$sdata_body .= hex2bin("00");           //[ 1]End of Username
		$sdata_body .= hex2bin("14");           //[ 1]Length of auth-response
		$sdata_body .= $pwd;
		$sdata_body .= "mysql_native_password"; //[21]payload
		$sdata_body .= hex2bin("00");           //[ 1]End of payload

		$sdata_body_len = strlen($sdata_body);
		$sdata = int2bin($sdata_body_len, 3);   //[ 3]Packet Length
		$sdata .= hex2bin("01");                //[ 1]Packet Number
		$sdata .= $sdata_body;                  //[ ?]Entire Packet

		pid_send($mysql_pid, $sdata, $sdata_body_len + 4);
		$mysql_state = MYSQL_STATE_AUTH;
		break;

	case MYSQL_STATE_AUTH:
		if(bin2int($rbuf, 4, 1) == 0)
		{
			echo "mysql: authentication is completed!\r\n";
			$mysql_state = MYSQL_STATE_QUERY_READY;
			return $mysql_pid;
		}
		else
		{
			echo "mysql: authentication failed!\r\n";
			mysql_close($mysql_pid);
		}
		break;
	}
	return false;
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_connect
// Description   : make progress MySQL connection phase
// Parameters    : $server - "IP address" or "IP:port" of DB Server
//               : $user_name - user account ID
//               : $raw_pwd - user account password
//               : $tcp_id - tcp id for $mysql_pid
// Return Values : false - on failure
//               : $pid - on sueccess; pid of MySQL connection
//////////////////////////////////////////////////////////////////////////
function mysql_connect($server, $user_name, $raw_pwd, $tcp_id = 3)
{
	mysql_init($server, $user_name, $raw_pwd);

	while(1)
	{
		$pid = mysql_con_loop($tcp_id);

		if($pid === false)
			usleep(1000);
		elseif($pid < 0)
			return false;
		else
			return $pid;
	}
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_query_loop
// Description   : make progress MySQL sending query phase in a loop
// Parameters    : none
// Return Values : false - for next call
//               : true - on success; by INSERT, UPDATE, DELETE and etc.
//               : $rbuf - on success; result by SELECT, SHOW and etc.
//               : -1 - on failure
//////////////////////////////////////////////////////////////////////////
function mysql_query_loop()
{
	global $mysql_pid;
	global $mysql_state;
	global $mysql_next_tick;
	global $mysql_query_str;

	$rbuf = "";

	if($mysql_state != MYSQL_STATE_QUERY_READY)
	{
		$len = pid_ioctl($mysql_pid, "get rxlen");
		if(!$len)
		{
			$state = pid_ioctl($mysql_pid, "get state");
			if($state != TCP_CONNECTED)
			{
				echo "mysql: connection closed!\r\n";
				return -1;
			}
			if(mysql_get_tick() >= $mysql_next_tick)
			{
				echo "mysql: receve timeout!\r\n";
				mysql_close($mysql_pid);
			}
			return false;
		}
		pid_recv($mysql_pid, $rbuf, $len);
	}

	switch($mysql_state)
	{
		case MYSQL_STATE_QUERY_READY:
			$q_data = hex2bin("03");
			$q_data .= $mysql_query_str;
			$slen = strlen($q_data);
			if(($slen > 0) && ($slen <= 0xff))
			{
				$q_sdata = int2bin($slen, 1);
				$q_sdata .= hex2bin("0000");
			}
			elseif(($slen >= 0xff) && ($slen <= 0xffff))
			{
				$q_sdata = int2bin($slen, 2);
				$q_sdata .= hex2bin("00");
			}
			elseif(($slen >= 0xffff) && ($slen <= 0xffffff))
				$q_sdata = int2bin($slen, 3);
			else
			{
				echo "mysql: invalid query!\r\n";
				return -1;
			}
			$q_sdata .= hex2bin("00");
			$q_sdata .= $q_data;
			pid_send($mysql_pid, $q_sdata);
			$mysql_next_tick = mysql_get_tick() + 5000;
			$mysql_state = MYSQL_STATE_QUERY;
			break;

		case MYSQL_STATE_QUERY:
			$mysql_state = MYSQL_STATE_QUERY_READY;
			if((bin2int($rbuf, 4, 1) == 0x00) && (bin2int($rbuf, 0, 3) <= 7))
				return true;
			return $rbuf;
	}

	return false;
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_query
// Description   : send a MySQL query 
// Parameters    : $query - query string for SQL
// Return Values : false - on failure
//               : true - on success; by INSERT, UPDATE, DELETE and etc.
//               : $ret - on success; result by SELECT, SHOW and etc.
//////////////////////////////////////////////////////////////////////////
function mysql_query($query)
{
	global $mysql_state;
	global $mysql_query_str;
	global $mysql_fetch_count, $mysql_text_pointer;

	$mysql_query_str = $query;
	$mysql_fetch_count = 0;
	$mysql_text_pointer = 0;

	if($mysql_state != MYSQL_STATE_QUERY_READY)
		return false;

	while(1)
	{
		$ret = mysql_query_loop();

		if($ret === false)
			usleep(1000);
		else
		{
			if($ret === true)
				return true;
			elseif($ret === -1)
				return false;
			else
				return $ret;
		}
	}
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_selectdb_loop
// Description   : make progress MySQL selecting db phase in a loop
// Parameters    : none
// Return Values : false - for next call
//               : true - on success; 
//               : $rbuf - on failure; response data from DB server
//               : -1 - on failure
//////////////////////////////////////////////////////////////////////////
function mysql_selectdb_loop()
{
	global $mysql_pid;
	global $mysql_state;
	global $mysql_next_tick;
	global $mysql_db_name;

	$rbuf = "";

	if($mysql_state != MYSQL_STATE_QUERY_READY)
	{
		$len = pid_ioctl($mysql_pid, "get rxlen");
		if(!$len)
		{
			$state = pid_ioctl($mysql_pid, "get state");
			if($state != TCP_CONNECTED)
			{
				echo "mysql: connection closed!\r\n";
				return -1;
			}
			if(mysql_get_tick() >= $mysql_next_tick)
			{
				echo "mysql: receve timeout!\r\n";
				mysql_close($mysql_pid);
			}
			return false;
		}
		pid_recv($mysql_pid, $rbuf, $len);
	}

	switch($mysql_state)
	{
		case MYSQL_STATE_QUERY_READY:
			$ud_data = hex2bin("02");      //Use Database Command
			$ud_data .= $mysql_db_name;    //Database Name
			$slen = strlen($ud_data);      //Length of Request
			$ud_sdata = int2bin($slen, 1); //Length of Request
			$ud_sdata .= hex2bin("0000");  //Padding
			$ud_sdata .= hex2bin("00");    //Packet Number
			$ud_sdata .= $ud_data;

			pid_send($mysql_pid, $ud_sdata);
			$mysql_next_tick = mysql_get_tick() + 5000;
			$mysql_state = MYSQL_STATE_QUERY;
			break;

		case MYSQL_STATE_QUERY:
			$mysql_state = MYSQL_STATE_QUERY_READY;
			if((bin2int($rbuf, 4, 1) == 0x00) && (bin2int($rbuf, 0, 3) <= 7))
				return true;
			return $rbuf;
	}

	return false;
}
//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_select_db
// Description   : select a MySQL database
// Parameters    : $db_name - DB name for usage
// Return Values : false - on failure
//               : true - on success;
//               : $ret - on failure; response data from DB server
//////////////////////////////////////////////////////////////////////////
function mysql_select_db($db_name)
{
	global $mysql_state;
	global $mysql_db_name;

	$mysql_db_name = $db_name;

	if($mysql_state != MYSQL_STATE_QUERY_READY)
		return false;

	while(1)
	{
		$ret = mysql_selectdb_loop();

		if($ret === false)
			usleep(1000);
		else
		{
			if($ret === true)
				return true;
			elseif($ret === -1)
				return false;
			else
				return $ret;
		}
	}
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_error
// Description   : get the text of the error message
// Parameters    : $result - received raw data from DB server
// Return Values : $error - the text of the error message
//////////////////////////////////////////////////////////////////////////
function mysql_error($result)
{
	$error = "";
	if(($result !== true) && ($result !== false))
	{
		$resp_head = bin2int($result, 4, 1);
		if($resp_head == 0xff)
			$error = substr($result,  13);
	}
	return $error;
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_errno
// Description   : get the numerical value of the error message
// Parameters    : $result - received raw data from DB server
// Return Values : $error - the numerical value of the error message
//////////////////////////////////////////////////////////////////////////
function mysql_errno($result)
{
	$errno = 0;
	if(($result !== true) && ($result !== false))
	{
		$resp_head = bin2int($result, 4, 1);
		if($resp_head == 0xff)
			$errno = bin2int($result, 5, 2);
	}
	return $errno;
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_affected_rows
// Description   : get number of affected rows
// Parameters    : $result - received raw data from DB server
// Return Values : -1 - on failure
//               : $aff_rwos - on success; number of affected rows
//////////////////////////////////////////////////////////////////////////
function mysql_affected_rows($result)
{
	$aff_rows = -1;
	if(($result !== true) && ($result !== false))
	{
		$resp_head = bin2int($result, 4, 1);
		if($resp_head == 0x00)
			$aff_rows = bin2int($result, 5, 1);
	}
	return $aff_rows;
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_num_rows
// Description   : get number of rows in result
// Parameters    : $result - received raw data from DB server
// Return Values : -1 - on failure
//               : $num_rwos - on success; number of rows
//////////////////////////////////////////////////////////////////////////
function mysql_num_rows($result)
{
	$num_rows = -1;
	if(($result !== true) && ($result !== false))
	{
		$resp_head = bin2int($result, 4, 1);

		if($resp_head != 0xff || $resp_head != 0xfe || $resp_head != 00)
			$num_cols = $resp_head;
		else
			return -1;
		$pos = strpos($result, int2bin(0xfe, 1)); 
		$num_1 = bin2int($result, $pos - 1, 1);
		$pos = strpos($result, int2bin(0xfe, 1), $pos + 1); 
		$num_2 = bin2int($result, $pos - 1, 1);
		$num_rows = $num_2 - $num_1 - 1;
	}
	return $num_rows;
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_result
// Description   : get result data
// Parameters    : $result - received raw data from DB server
//               : row - the row number
//               : col - the column number 
// Return Values : false - on failure
//               : $text - on success; the contents of one cell
//////////////////////////////////////////////////////////////////////////
function mysql_result($result, $row, $col = 0)
{
	if(($result === true) || ($result === false))
		return false;

	$num_cols = bin2int($result, 4, 1);
	if($num_cols < $col)
		return false;

	$num_rows = mysql_num_rows($result);
	if($num_rows < $row)
		return false;

	$pointer = strpos($result, int2bin(0xfe, 1)) + 5;
	//skipping records
	for($i = 0; $i < $row; $i++)
	{
		$packet_len = bin2int($result, $pointer, 3);
		$pointer += ($packet_len + 4);
	}
	$pointer += 4;
	//skipping columns
	for($i = 0; $i < $col; $i++)
	{
		$param_len = bin2int($result, $pointer, 1);
		$pointer += ($param_len + 1);
	}
	$param_len = bin2int($result, $pointer, 1);
	$text = substr($result, $pointer + 1, $param_len);
	return $text;
}

//////////////////////////////////////////////////////////////////////////
// Function Name : mysql_fetch_row
// Description   : get a result row as an enumerated array
// Parameters    : $result - received raw data from DB server
// Return Values : false - on failure
//               : $result_arr - on success; evaluated result row 
//////////////////////////////////////////////////////////////////////////
function mysql_fetch_row($result)
{
	global $mysql_fetch_count;
	global $mysql_text_pointer;

	$result_arr = array("", "", "", "", "", "", "", "");

	$num_cols = bin2int($result, 4, 1);
	$num_rows = mysql_num_rows($result);

	if($mysql_fetch_count>= $num_rows)
		return false;

	if($mysql_fetch_count== 0)
		$mysql_text_pointer = strpos($result, int2bin(0xfe, 1)) + 5;

	$packet_len = bin2int($result, $mysql_text_pointer, 3);
	$mysql_text_pointer += 4;

	for($i = 0; $i < $num_cols; $i++)
	{
		$len = bin2int($result, $mysql_text_pointer, 1);
		$result_arr[$i] = substr($result, $mysql_text_pointer + 1, $len);
		$mysql_text_pointer += ($len + 1);
	}
	
	$mysql_fetch_count++;
	return $result_arr;
}


?>

